# pylint: disable=unused-import,missing-class-docstring
import logging
import os
import requests
import time

from tlbx import (
    st,
    dbg as _dbg,
    dbgsql as _dbgsql,
    info as _info,
    warn as _warn,
    error as _error,
    pf,
    pp,
    format_msg,
    sqlformat,
    rmfile,
    get_caller,
    open_filepath_or_buffer,
    get_string_format_args,
    json,
    chunks,
    iter_or,
    powerset,
    is_int,
    orderedsetify,
    initializer,
    PrintMixin,
    MappingMixin,
    ClassValueContainsMeta,
    get_class_var_values,
    get_class_vars,
    import_object,
)
import yaml


ADHOC_DS_URL = "adhoc"  # A placeholder to denote its an adhoc datasource
RESERVED_FIELD_NAMES = set(["row_hash"])
DEFAULT_REPLACE_AFTER = "1 days"
CRITERIA_OPERATIONS = set(
    [
        ">",
        ">=",
        "<",
        "<=",
        "=",
        "!=",
        "in",
        "not in",
        "between",
        "not between",
        "like",
        "not like",
    ]
)
ROW_FILTER_OPERATIONS = set([">", ">=", "<", "<=", "=", "!="])


default_logger = logging.getLogger("zillion")
default_logger.setLevel(logging.INFO)


def dbg(msg, **kwargs):
    """Call tlbx dbg with zillion logger"""
    kwargs["logger"] = kwargs.get("logger", default_logger)
    kwargs["label"] = kwargs.get("label", get_caller())
    _dbg(msg, **kwargs)


def dbgsql(msg, **kwargs):
    """Call tlbx dbgsql with zillion logger"""
    kwargs["logger"] = kwargs.get("logger", default_logger)
    kwargs["label"] = kwargs.get("label", get_caller())
    _dbgsql(msg, **kwargs)


def info(msg, **kwargs):
    """Call tlbx info with zillion logger"""
    kwargs["logger"] = kwargs.get("logger", default_logger)
    kwargs["label"] = kwargs.get("label", get_caller())
    _info(msg, **kwargs)


def warn(msg, **kwargs):
    """Call tlbx warn with zillion logger"""
    kwargs["logger"] = kwargs.get("logger", default_logger)
    kwargs["label"] = kwargs.get("label", get_caller())
    _warn(msg, **kwargs)


def error(msg, **kwargs):
    """Call tlbx error with zillion logger"""
    kwargs["logger"] = kwargs.get("logger", default_logger)
    kwargs["label"] = kwargs.get("label", get_caller())
    _error(msg, **kwargs)


class ZillionException(Exception):
    pass


class InvalidTechnicalException(ZillionException):
    pass


class WarehouseException(ZillionException):
    pass


class ReportException(ZillionException):
    pass


class InvalidReportIdException(ZillionException):
    pass


class UnsupportedGrainException(ZillionException):
    pass


class UnsupportedKillException(ZillionException):
    pass


class FailedKillException(ZillionException):
    pass


class DataSourceQueryTimeoutException(ZillionException):
    pass


class ExecutionKilledException(ZillionException):
    pass


class ExecutionLockException(ZillionException):
    pass


class InvalidFieldException(ZillionException):
    pass


class InvalidDimensionValueException(ZillionException):
    pass


class DisallowedSQLException(ZillionException):
    pass


class MaxFormulaDepthException(ZillionException):
    pass


class FieldTypes(metaclass=ClassValueContainsMeta):
    """Allowed field types"""

    DIMENSION = "dimension"
    METRIC = "metric"


class TableTypes(metaclass=ClassValueContainsMeta):
    """Allowed table types"""

    DIMENSION = "dimension"
    METRIC = "metric"


class AggregationTypes(metaclass=ClassValueContainsMeta):
    """Allowed aggregation types. These aggregations are limited by what can be
    done in most SQL databases"""

    MEAN = "mean"
    SUM = "sum"
    MIN = "min"
    MAX = "max"
    COUNT = "count"
    COUNT_DISTINCT = "count_distinct"


class TechnicalTypes(metaclass=ClassValueContainsMeta):
    """Allowed technical types"""

    MEAN = "mean"
    SUM = "sum"
    MEDIAN = "median"
    STD = "std"
    VAR = "var"
    MIN = "min"
    MAX = "max"
    BOLL = "boll"
    DIFF = "diff"
    PCT_CHANGE = "pct_change"
    CUMSUM = "cumsum"
    CUMMIN = "cummin"
    CUMMAX = "cummax"
    RANK = "rank"
    PCT_RANK = "pct_rank"


class TechnicalModes(metaclass=ClassValueContainsMeta):
    """Allowed Technical modes
    
    **Attributes:**
    
    * **GROUP** - (*str*) Apply the technical to the last grouping of the data
    for a multi-dimensional report
    * **ALL** - (*str*) Apply the technical across all result data
    
    """

    GROUP = "group"
    ALL = "all"


class RollupTypes(metaclass=ClassValueContainsMeta):
    """Allowed Rollup Types"""

    TOTALS = "totals"
    ALL = "all"


class OrderByTypes(metaclass=ClassValueContainsMeta):
    """Allowed Order By Types"""

    ASC = "asc"
    DESC = "desc"


class DataSourceQueryModes(metaclass=ClassValueContainsMeta):
    """Allowed datasource query modes"""

    SEQUENTIAL = "sequential"
    MULTITHREAD = "multithread"


class ExecutionState:
    """Allowed report/query execution states"""

    READY = "ready"
    QUERYING = "querying"
    KILLED = "killed"


class IfExistsModes(metaclass=ClassValueContainsMeta):
    """Allowed modes when creating tables from data. This is based off of pandas
    `if_exists` param in the `DataFrame.to_sql` method, with the addition of an
    "ignore" option. The "append" option is also removed for now since there
    isn't a safe/generic way to guarantee a proper primary key has been set on
    the table."""

    FAIL = "fail"
    REPLACE = "replace"
    # APPEND = "append"
    IGNORE = "ignore"


class IfFileExistsModes(IfExistsModes):
    """An extension of the modes above specific to downloaded files. This
    allows the config to specify that a downloaded file should be replaced
    after a certain amount of time. See code that uses this for implementation
    details.
    """

    REPLACE_AFTER = "replace_after"


# ---- TODO: move below to utils file


def raiseif(cond, msg="", exc=ZillionException):
    """Convenience assert-like utility"""
    if cond:
        raise exc(msg)


def raiseifnot(cond, msg="", exc=ZillionException):
    """Convenience assert-like utility"""
    if not cond:
        raise exc(msg)


def igetattr(obj, attr, *args):
    """Case-insensitive getattr"""
    for a in dir(obj):
        if a.lower() == attr.lower():
            return getattr(obj, a)
    if args:
        return args[0]
    raise AttributeError("type object '%s' has no attribute '%s'" % (type(obj), attr))


def read_filepath_or_buffer(f, open_flags="r", compression=None):
    """Open and read files or buffers, local or remote"""
    f, handles, close = open_filepath_or_buffer(
        f, open_flags=open_flags, compression=compression
    )
    try:
        data = f.read()
    finally:
        if close:
            try:
                f.close()
            except ValueError:
                pass
    return data


def download_file(url, outfile=None):
    """Utility to download a datafile"""
    if not outfile:
        outfile = url.split("/")[-1]
    info("Downloading %s to %s" % (url, outfile))
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(outfile, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return outfile


def get_modified_time(fname):
    """Utility to get the modified time of a file"""
    return os.stat(fname).st_mtime


def get_time_since_modified(fname):
    """Utility to get the time since a file was last modified"""
    return time.time() - get_modified_time(fname)


def load_yaml(fname):
    """Wrapper to safe_load that also expands environment vars"""
    with open(fname) as f:
        val = yaml.safe_load(os.path.expandvars(f.read()))
    return val



"""Zillion package"""

from .version import __version__
from .core import (
    FieldTypes,
    TableTypes,
    AggregationTypes,
    TechnicalTypes,
    TechnicalModes,
    RollupTypes,
    DataSourceQueryModes,
    IfExistsModes,
)
from .configs import (
    load_zillion_config,
    load_warehouse_config,
    load_warehouse_config_from_env,
    load_datasource_config,
    load_datasource_config_from_env,
)
from .datasource import DataSource
from .warehouse import Warehouse
from .report import Report



POSTGRESQL_YEAR_CRITERIA_CONVERSIONS = {
    "=": [
        [">=", "TO_DATE(:0 || '-01-01', 'YYYY-MM-DD')"],
        ["<", "(TO_DATE(:0 || '-01-01', 'YYYY-MM-DD') + interval '1 year')"],
    ],
    # NOTE: all of the conditions in the criteria replacement lists get AND'd together
    # so we can't simply do "x < 2020-01-01 or x >= 2021-01-01"
    "!=": [
        [
            "not between",
            [
                "TO_DATE(:0 || '-01-01', 'YYYY-MM-DD')",
                "(TO_DATE(:0 || '-01-01', 'YYYY-MM-DD') + interval '1 year' - interval '1 second')",
            ],
        ]
    ],
    ">": [[">=", "(TO_DATE(:0 || '-01-01', 'YYYY-MM-DD') + interval '1 year')"]],
    ">=": [[">=", "TO_DATE(:0 || '-01-01', 'YYYY-MM-DD')"]],
    "<": [["<", "TO_DATE(:0 || '-01-01', 'YYYY-MM-DD')"]],
    "<=": [["<", "(TO_DATE(:0 || '-01-01', 'YYYY-MM-DD') + interval '1 year')"]],
    "between": [
        [">=", "TO_DATE(:0 || '-01-01', 'YYYY-MM-DD')"],
        ["<", "(TO_DATE(:1 || '-01-01', 'YYYY-MM-DD') + interval '1 year')"],
    ],
    "not between": [
        [
            "not between",
            [
                "TO_DATE(:0 || '-01-01', 'YYYY-MM-DD')",
                "(TO_DATE(:1 || '-01-01', 'YYYY-MM-DD') + interval '1 year' - interval '1 second')",
            ],
        ]
    ],
}

POSTGRESQL_MONTH_CRITERIA_CONVERSIONS = {
    "=": [
        [">=", "TO_DATE(:0 || '-01', 'YYYY-MM')"],
        ["<", "(TO_DATE(:0 || '-01', 'YYYY-MM') + interval '1 month')"],
    ],
    "!=": [
        [
            "not between",
            [
                "TO_DATE(:0 || '-01', 'YYYY-MM')",
                "(TO_DATE(:0 || '-01', 'YYYY-MM') + interval '1 month' - interval '1 second')",
            ],
        ]
    ],
    ">": [[">=", "(TO_DATE(:0 || '-01', 'YYYY-MM') + interval '1 month')"]],
    ">=": [[">=", "TO_DATE(:0 || '-01', 'YYYY-MM')"]],
    "<": [["<", "TO_DATE(:0 || '-01', 'YYYY-MM')"]],
    "<=": [["<", "(TO_DATE(:0 || '-01', 'YYYY-MM') + interval '1 month')"]],
    "between": [
        [">=", "TO_DATE(:0 || '-01', 'YYYY-MM')"],
        ["<", "(TO_DATE(:0 || '-01', 'YYYY-MM') + interval '1 month')"],
    ],
    "not between": [
        [
            "not between",
            [
                "TO_DATE(:0 || '-01', 'YYYY-MM')",
                "(TO_DATE(:0 || '-01', 'YYYY-MM') + interval '1 month' - interval '1 second')",
            ],
        ]
    ],
}

POSTGRESQL_DATE_CRITERIA_CONVERSIONS = {
    "=": [[">=", ":0"], ["<", "(TO_DATE(:0, 'YYYY-MM-DD') + interval '1 day')"]],
    "!=": [
        [
            "not between",
            [
                ":0",
                "(TO_DATE(:0, 'YYYY-MM-DD') + interval '1 day' - interval '1 second')",
            ],
        ]
    ],
    ">": [[">=", "(TO_DATE(:0, 'YYYY-MM-DD') + interval '1 day')"]],
    ">=": [[">=", ":0"]],
    "<": [["<", ":0"]],
    "<=": [["<", "(TO_DATE(:0, 'YYYY-MM-DD') + interval '1 day')"]],
    "between": [[">=", ":0"], ["<", "(TO_DATE(:1, 'YYYY-MM-DD') + interval '1 day')"]],
    "not between": [
        [
            "not between",
            [
                ":0",
                "(TO_DATE(:1, 'YYYY-MM-DD') + interval '1 day' - interval '1 second')",
            ],
        ]
    ],
}

POSTGRESQL_HOUR_CRITERIA_CONVERSIONS = {
    "=": [
        [">=", ":0"],
        ["<", "(TO_TIMESTAMP(:0, 'YYYY-MM-DD HH24:MI:SS') + interval '1 hour')"],
    ],
    "!=": [
        [
            "not between",
            [
                ":0",
                "(TO_TIMESTAMP(:0, 'YYYY-MM-DD HH24:MI:SS') + interval '1 hour' - interval '1 second')",
            ],
        ]
    ],
    ">": [[">=", "(TO_TIMESTAMP(:0, 'YYYY-MM-DD HH24:MI:SS') + interval '1 hour')"]],
    ">=": [[">=", ":0"]],
    "<": [["<", ":0"]],
    "<=": [["<", "(TO_TIMESTAMP(:0, 'YYYY-MM-DD HH24:MI:SS') + interval '1 hour')"]],
    "between": [
        [">=", ":0"],
        ["<", "(TO_TIMESTAMP(:1, 'YYYY-MM-DD HH24:MI:SS') + interval '1 hour')"],
    ],
    "not between": [
        [
            "not between",
            [
                ":0",
                "(TO_TIMESTAMP(:1, 'YYYY-MM-DD HH24:MI:SS') + interval '1 hour' - interval '1 second')",
            ],
        ]
    ],
}

POSTGRESQL_MINUTE_CRITERIA_CONVERSIONS = {
    "=": [
        [">=", ":0"],
        ["<", "(TO_TIMESTAMP(:0, 'YYYY-MM-DD HH24:MI:SS') + interval '1 minute')"],
    ],
    "!=": [
        [
            "not between",
            [
                ":0",
                "(TO_TIMESTAMP(:0, 'YYYY-MM-DD HH24:MI:SS') + interval '1 minute' - interval '1 second')",
            ],
        ]
    ],
    ">": [[">=", "(TO_TIMESTAMP(:0, 'YYYY-MM-DD HH24:MI:SS') + interval '1 minute')"]],
    ">=": [[">=", ":0"]],
    "<": [["<", ":0"]],
    "<=": [["<", "(TO_TIMESTAMP(:0, 'YYYY-MM-DD HH24:MI:SS') + interval '1 minute')"]],
    "between": [
        [">=", ":0"],
        ["<", "(TO_TIMESTAMP(:1, 'YYYY-MM-DD HH24:MI:SS') + interval '1 minute')"],
    ],
    "not between": [
        [
            "not between",
            [
                ":0",
                "(TO_TIMESTAMP(:1, 'YYYY-MM-DD HH24:MI:SS') + interval '1 minute' - interval '1 second')",
            ],
        ]
    ],
}

POSTGRESQL_DATETIME_CRITERIA_CONVERSIONS = {
    "=": [["=", ":0"]],
    "!=": [["!=", ":0"]],
    ">": [[">", ":0"]],
    ">=": [[">=", ":0"]],
    "<": [["<", ":0"]],
    "<=": [["<=", ":0"]],
    "between": [["between", [":0", ":1"]]],
    "not between": [["not between", [":0", ":1"]]],
}

POSTGRESQL_DIALECT_CONVERSIONS = {
    "year": {
        "ds_formula": "EXTRACT(YEAR FROM {})",
        "ds_criteria_conversions": POSTGRESQL_YEAR_CRITERIA_CONVERSIONS,
    },
    "quarter": "TO_CHAR({}, 'FMYYYY-\"Q\"Q')",
    "quarter_of_year": "EXTRACT(QUARTER FROM {})",
    "month": {
        "ds_formula": "TO_CHAR({}, 'FMYYYY-MM')",
        "ds_criteria_conversions": POSTGRESQL_MONTH_CRITERIA_CONVERSIONS,
    },
    "month_name": "TO_CHAR({}, 'FMMonth')",
    "month_of_year": "EXTRACT(MONTH FROM {})",
    "date": {
        "ds_formula": "TO_CHAR({}, 'FMYYYY-MM-DD')",
        "ds_criteria_conversions": POSTGRESQL_DATE_CRITERIA_CONVERSIONS,
    },
    "day_name": "TO_CHAR({}, 'FMDay')",
    "day_of_week": "EXTRACT(ISODOW FROM {})",  # Monday = 1
    "day_of_month": "EXTRACT(DAY FROM {})",
    "day_of_year": "EXTRACT(DOY FROM {})",
    "hour": {
        "ds_formula": "TO_CHAR({}, 'FMYYYY-MM-DD HH24:00:00')",
        "ds_criteria_conversions": POSTGRESQL_HOUR_CRITERIA_CONVERSIONS,
    },
    "hour_of_day": "EXTRACT(HOUR FROM {})",
    "minute": {
        "ds_formula": "TO_CHAR({}, 'FMYYYY-MM-DD HH24:MI:00')",
        "ds_criteria_conversions": POSTGRESQL_MINUTE_CRITERIA_CONVERSIONS,
    },
    "minute_of_hour": "EXTRACT(MINUTE FROM {})",
    "datetime": {
        "ds_formula": "TO_CHAR({}, 'FMYYYY-MM-DD HH24:MI:SS')",
        "ds_criteria_conversions": POSTGRESQL_DATETIME_CRITERIA_CONVERSIONS,
    },
    "unixtime": "EXTRACT(epoch from {})",
}



MYSQL_YEAR_CRITERIA_CONVERSIONS = {
    "=": [
        [">=", "CONCAT(:0, '-01-01')"],
        ["<", "DATE_ADD(CONCAT(:0, '-01-01'), INTERVAL 1 YEAR)"],
    ],
    # NOTE: all of the conditions in the criteria replacement lists get AND'd together
    # so we can't simply do "x < 2020-01-01 or x >= 2021-01-01"
    "!=": [
        [
            "not between",
            [
                "CONCAT(:0, '-01-01')",
                "DATE_SUB(DATE_ADD(CONCAT(:0, '-01-01'), INTERVAL 1 YEAR), INTERVAL 1 SECOND)",
            ],
        ]
    ],
    ">": [[">=", "DATE_ADD(CONCAT(:0, '-01-01'), INTERVAL 1 YEAR)"]],
    ">=": [[">=", "CONCAT(:0, '-01-01')"]],
    "<": [["<", "CONCAT(:0, '-01-01')"]],
    "<=": [["<", "DATE_ADD(CONCAT(:0, '-01-01'), INTERVAL 1 YEAR)"]],
    "between": [
        [">=", "CONCAT(:0, '-01-01')"],
        ["<", "DATE_ADD(CONCAT(:1, '-01-01'), INTERVAL 1 YEAR)"],
    ],
    "not between": [
        [
            "not between",
            [
                "CONCAT(:0, '-01-01')",
                "DATE_SUB(DATE_ADD(CONCAT(:1, '-01-01'), INTERVAL 1 YEAR), INTERVAL 1 SECOND)",
            ],
        ]
    ],
}

MYSQL_MONTH_CRITERIA_CONVERSIONS = {
    "=": [
        [">=", "CONCAT(:0, '-01')"],
        ["<", "DATE_ADD(CONCAT(:0, '-01'), INTERVAL 1 MONTH)"],
    ],
    "!=": [
        [
            "not between",
            [
                "CONCAT(:0, '-01')",
                "DATE_SUB(DATE_ADD(CONCAT(:0, '-01'), INTERVAL 1 MONTH), INTERVAL 1 SECOND)",
            ],
        ]
    ],
    ">": [[">=", "DATE_ADD(CONCAT(:0, '-01'), INTERVAL 1 MONTH)"]],
    ">=": [[">=", "CONCAT(:0, '-01')"]],
    "<": [["<", "CONCAT(:0, '-01')"]],
    "<=": [["<", "DATE_ADD(CONCAT(:0, '-01'), INTERVAL 1 MONTH)"]],
    "between": [
        [">=", "CONCAT(:0, '-01')"],
        ["<", "DATE_ADD(CONCAT(:1, '-01'), INTERVAL 1 MONTH)"],
    ],
    "not between": [
        [
            "not between",
            [
                "CONCAT(:0, '-01')",
                "DATE_SUB(DATE_ADD(CONCAT(:1, '-01'), INTERVAL 1 MONTH), INTERVAL 1 SECOND)",
            ],
        ]
    ],
}

MYSQL_DATE_CRITERIA_CONVERSIONS = {
    "=": [[">=", ":0"], ["<", "DATE_ADD(:0, INTERVAL 1 DAY)"]],
    "!=": [
        [
            "not between",
            [":0", "DATE_SUB(DATE_ADD(:0, INTERVAL 1 DAY), INTERVAL 1 SECOND)"],
        ]
    ],
    ">": [[">=", "DATE_ADD(:0, INTERVAL 1 DAY)"]],
    ">=": [[">=", ":0"]],
    "<": [["<", ":0"]],
    "<=": [["<", "DATE_ADD(:0, INTERVAL 1 DAY)"]],
    "between": [[">=", ":0"], ["<", "DATE_ADD(:1, INTERVAL 1 DAY)"]],
    "not between": [
        [
            "not between",
            [":0", "DATE_SUB(DATE_ADD(:1, INTERVAL 1 DAY), INTERVAL 1 SECOND)"],
        ]
    ],
}

MYSQL_HOUR_CRITERIA_CONVERSIONS = {
    "=": [[">=", ":0"], ["<", "DATE_ADD(:0, INTERVAL 1 HOUR)"]],
    "!=": [
        [
            "not between",
            [":0", "DATE_SUB(DATE_ADD(:0, INTERVAL 1 HOUR), INTERVAL 1 SECOND)"],
        ]
    ],
    ">": [[">=", "DATE_ADD(:0, INTERVAL 1 HOUR)"]],
    ">=": [[">=", ":0"]],
    "<": [["<", ":0"]],
    "<=": [["<", "DATE_ADD(:0, INTERVAL 1 HOUR)"]],
    "between": [[">=", ":0"], ["<", "DATE_ADD(:1, INTERVAL 1 HOUR)"]],
    "not between": [
        [
            "not between",
            [":0", "DATE_SUB(DATE_ADD(:1, INTERVAL 1 HOUR), INTERVAL 1 SECOND)"],
        ]
    ],
}

MYSQL_MINUTE_CRITERIA_CONVERSIONS = {
    "=": [[">=", ":0"], ["<", "DATE_ADD(:0, INTERVAL 1 MINUTE)"]],
    "!=": [
        [
            "not between",
            [":0", "DATE_SUB(DATE_ADD(:0, INTERVAL 1 MINUTE), INTERVAL 1 SECOND)"],
        ]
    ],
    ">": [[">=", "DATE_ADD(:0, INTERVAL 1 MINUTE)"]],
    ">=": [[">=", ":0"]],
    "<": [["<", ":0"]],
    "<=": [["<", "DATE_ADD(:0, INTERVAL 1 MINUTE)"]],
    "between": [[">=", ":0"], ["<", "DATE_ADD(:1, INTERVAL 1 MINUTE)"]],
    "not between": [
        [
            "not between",
            [":0", "DATE_SUB(DATE_ADD(:1, INTERVAL 1 MINUTE), INTERVAL 1 SECOND)"],
        ]
    ],
}

MYSQL_DATETIME_CRITERIA_CONVERSIONS = {
    "=": [["=", ":0"]],
    "!=": [["!=", ":0"]],
    ">": [[">", ":0"]],
    ">=": [[">=", ":0"]],
    "<": [["<", ":0"]],
    "<=": [["<=", ":0"]],
    "between": [["between", [":0", ":1"]]],
    "not between": [["not between", [":0", ":1"]]],
}

MYSQL_DIALECT_CONVERSIONS = {
    "year": {
        "ds_formula": "EXTRACT(YEAR FROM {})",
        "ds_criteria_conversions": MYSQL_YEAR_CRITERIA_CONVERSIONS,
    },
    "quarter": "CONCAT(YEAR({}), '-Q', QUARTER({}))",
    "quarter_of_year": "EXTRACT(QUARTER FROM {})",
    "month": {
        "ds_formula": "DATE_FORMAT({}, '%Y-%m')",
        "ds_criteria_conversions": MYSQL_MONTH_CRITERIA_CONVERSIONS,
    },
    "month_name": "MONTHNAME({})",
    "month_of_year": "EXTRACT(MONTH FROM {})",
    "date": {
        "ds_formula": "DATE_FORMAT({}, '%Y-%m-%d')",
        "ds_criteria_conversions": MYSQL_DATE_CRITERIA_CONVERSIONS,
    },
    "day_name": "DAYNAME({})",
    "day_of_week": "WEEKDAY({}) + 1",  # Monday = 1
    "day_of_month": "EXTRACT(DAY FROM {})",
    "day_of_year": "DAYOFYEAR({})",
    "hour": {
        "ds_formula": "DATE_FORMAT({}, '%Y-%m-%d %H:00:00')",
        "ds_criteria_conversions": MYSQL_HOUR_CRITERIA_CONVERSIONS,
    },
    "hour_of_day": "EXTRACT(HOUR FROM {})",
    "minute": {
        "ds_formula": "DATE_FORMAT({}, '%Y-%m-%d %H:%i:00')",
        "ds_criteria_conversions": MYSQL_MINUTE_CRITERIA_CONVERSIONS,
    },
    "minute_of_hour": "EXTRACT(MINUTE FROM {})",
    "datetime": {
        "ds_formula": "DATE_FORMAT({}, '%Y-%m-%d %H:%i:%S')",
        "ds_criteria_conversions": MYSQL_DATETIME_CRITERIA_CONVERSIONS,
    },
    "unixtime": "UNIX_TIMESTAMP({})",
}



import pandas as pd
import sqlalchemy as sa

from zillion.configs import (
    EXCLUDE,
    ConfigMixin,
    FieldConfigSchema,
    FormulaFieldConfigSchema,
    MetricConfigSchema,
    FormulaMetricConfigSchema,
    DimensionConfigSchema,
    AdHocMetricSchema,
    AdHocFieldSchema,
    create_technical,
    is_valid_field_name,
    is_active,
    default_field_display_name,
)
from zillion.core import *
from zillion.dialects import *
from zillion.model import zillion_engine, DimensionValues
from zillion.sql_utils import (
    aggregation_to_sqla_func,
    contains_aggregation,
    contains_sql_keywords,
    type_string_to_sa_type,
    to_generic_sa_type,
    sqla_compile,
    column_fullname,
)


MAX_FORMULA_DEPTH = 3
# This default warehouse ID is used if no ID has been populated
# on the Warehouse when checking dimension values.
FIELD_VALUE_DEFAULT_WAREHOUSE_ID = 0
FIELD_VALUE_CHECK_OPERATIONS = set(["=", "!=", "in", "not in"])


class Field(ConfigMixin, PrintMixin):
    """Represents the concept a column is capturing, which may be shared by
    columns in other tables or datasources. For example, you may have a several
    columns in your databases/tables that represent the concept of "revenue". In
    other words, a column is like an instance of a Field.
    
    **Parameters:**
    
    * **name** - (*str*) The name of the field
    * **type** - (*str or SQLAlchemy type*) The column type for the field.
    * **display_name** - (*str, optional*) The display name of the field
    * **description** - (*str, optional*) The description of the field
    * **meta** - (*dict, optional*) A dict of additional custom attributes
    * **kwargs** - Additional attributes stored on the field object
    
    **Attributes:**
    
    * **name** - (*str*) The name of the field
    * **type** - (*str*) A string representing the generic SQLAlchemy type
    * **display_name** - (*str, optional*) The display name of the field
    * **description** - (*str, optional*) The description of the field
    * **meta** - (*dict, optional*) A dict of additional custom attributes
    * **sa_type** - (*SQLAlchemy type*) If a dialect-specific type object
    is passed in on init it will be coerced to a generic type.
    * **field_type** - (*str*) A valid FieldType string
    * **schema** - (*Marshmallow schema*) A FieldConfigSchema class
    
    """

    repr_attrs = ["name", "type"]
    field_type = None
    schema = FieldConfigSchema

    @initializer
    def __init__(
        self, name, type, display_name=None, description=None, meta=None, **kwargs
    ):
        self.sa_type = None
        if isinstance(type, str):
            self.sa_type = type_string_to_sa_type(type)
        elif type:
            self.sa_type = to_generic_sa_type(type)
            self.type = repr(self.sa_type)
        # This will do schema validation
        super().__init__()

    def copy(self):
        """Copy this field"""
        return self.__class__.from_config(self.to_config())

    def get_formula_fields(self, warehouse, depth=0, adhoc_fms=None):
        """Get the fields that are part of this field's formula
        
        **Parameters:**
        
        * **warehouse** - (*Warehouse*) A zillion warehouse that will contain
        all relevant fields
        * **depth** - (*int, optional*) Track the depth of recursion into the
        formula
        * **adhoc_fms** - (*list, optional*) A list of FieldManagers
        
        **Returns:**
        
        (*set, str*) - The set of all base fields involved in the formula
        calculation, as well as an expanded version of the formula. All fields
        in the expanded formula should be raw fields (i.e. not formula
        fields).
        
        """
        return None, None

    def get_ds_expression(self, column, label=True, ignore_formula=False):
        """Get the datasource-level sql expression for this field
        
        **Parameters:**
        
        * **column** - (*Column*) A SQLAlchemy column that supports this field
        * **label** - (*bool, optional*) If true, label the expression with the
        field name
        * **ignore_formula** - (*bool, optional*) If true, don't apply any available
        datasource formulas
        
        """
        ds_formula = column.zillion.field_ds_formula(self.name)
        if ignore_formula or (not ds_formula):
            if label:
                return column.label(self.name)
            return column

        if contains_sql_keywords(ds_formula):
            raise DisallowedSQLException(
                "Formula contains disallowed sql: %s" % ds_formula
            )

        if not (ds_formula.startswith("(") and ds_formula.endswith("(")):
            ds_formula = "(" + ds_formula + ")"

        if label:
            return sa.literal_column(ds_formula).label(self.name)
        return sa.literal_column(ds_formula)

    def get_final_select_clause(self, *args, **kwargs):
        """The sql clause used when selecting at the combined query layer"""
        return self.name

    # https://stackoverflow.com/questions/2909106/whats-a-correct-and-good-way-to-implement-hash

    def __key(self):
        return self.name

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(self, type(other)) and self.__key() == other.__key()


class Metric(Field):
    """Fields that represent values to be measured and possibly broken down
    along Dimensions
    
    **Parameters:**
    
    * **name** - (*str*) The name of the field
    * **type** - (*str or SQLAlchemy type*) The column type for the field
    * **display_name** - (*str, optional*) The display name of the field
    * **description** - (*str, optional*) The description of the field
    * **meta** - (*dict, optional*) A dict of additional custom attributes
    * **aggregation** - (*str, optional*) The AggregationType to apply to the
    metric
    * **rounding** - (*int, optional*) If specified, the number of decimal
    places to round to
    * **weighting_metric** - (*str, optional*) A reference to a metric to use
    for weighting when aggregating averages
    * **technical** - (*object, optional*) A Technical object or definition used
    to defined a technical computation to be applied to the metric
    * **required_grain** - (*list of str, optional*) If specified, a list of
    dimensions that must be present in the dimension grain of any report that
    aims to include this metric.
    * **kwargs** - kwargs passed to super class
    
    """

    field_type = FieldTypes.METRIC
    schema = MetricConfigSchema

    def __init__(
        self,
        name,
        type,
        display_name=None,
        description=None,
        meta=None,
        aggregation=AggregationTypes.SUM,
        rounding=None,
        weighting_metric=None,
        technical=None,
        required_grain=None,
        **kwargs
    ):
        if weighting_metric:
            raiseifnot(
                aggregation == AggregationTypes.MEAN,
                'Weighting metrics are only supported for "%s" aggregation type'
                % AggregationTypes.MEAN,
            )

        if technical:
            technical = create_technical(technical)

        super(Metric, self).__init__(
            name,
            type,
            display_name=display_name,
            description=description,
            meta=meta,
            aggregation=aggregation,
            rounding=rounding,
            weighting_metric=weighting_metric,
            technical=technical,
            required_grain=required_grain,
            **kwargs
        )

    def get_ds_expression(self, column, label=True):
        """Get the datasource-level sql expression for this metric
        
        **Parameters:**
        
        * **column** - (*Column*) A SQLAlchemy column that supports this metric
        * **label** - (*bool, optional*) If true, label the expression with the
        field name
        
        """
        expr = column
        aggr = aggregation_to_sqla_func(self.aggregation)
        skip_aggr = False

        ds_formula = column.zillion.field_ds_formula(self.name)

        if ds_formula:
            if contains_sql_keywords(ds_formula):
                raise DisallowedSQLException(
                    "Formula contains disallowed sql: %s" % ds_formula
                )
            if contains_aggregation(ds_formula):
                info("Datasource formula contains aggregation, skipping default logic")
                skip_aggr = True
            expr = sa.literal_column(ds_formula)

        if not skip_aggr:
            if self.aggregation in [
                AggregationTypes.COUNT,
                AggregationTypes.COUNT_DISTINCT,
            ]:
                if self.rounding:
                    info("Ignoring rounding for count field: %s" % self.name)
                if label:
                    return aggr(expr).label(self.name)
                return aggr(expr)

            if self.weighting_metric:
                w_column = get_table_field_column(column.table, self.weighting_metric)
                w_column_name = column_fullname(w_column)
                # NOTE: 1.0 multiplication is a hack to ensure results are not rounded
                # to integer values improperly by some database dialects such as sqlite
                expr = sa.func.SUM(
                    sa.text("1.0") * expr * sa.text(w_column_name)
                ) / sa.func.SUM(sa.text(w_column_name))
            else:
                expr = aggr(expr)

        if label:
            return expr.label(self.name)
        return expr

    def get_final_select_clause(self, *args, **kwargs):
        """The sql clause used when selecting at the combined query layer"""
        return self.name


class Dimension(Field):
    """Fields that represent attributes of data that are used for grouping or
    filtering
    
    **Parameters:**
    
    * **name** - (*str*) The name of the field
    * **type** - (*str or SQLAlchemy type*) The column type for the field.
    * **display_name** - (*str, optional*) The display name of the field
    * **description** - (*str, optional*) The description of the field
    * **values** - (*str or list, optional*) A list of allowed dimension
    values or a name of a callable to provide a list of values
    * **sorter** - (*str, optional*) A reference to an importable callable
    that accepts three arguments: (warehouse ID, dimension object, values).
    Currently values is a pandas Series and the callable is expected to
    return a Series. See `zillion.field.sort_by_value_order` for an example.
    * **meta** - (*dict, optional*) A dict of additional custom attributes
    * **kwargs** - Additional attributes stored on the field object
    
    """

    field_type = FieldTypes.DIMENSION
    schema = DimensionConfigSchema

    @initializer
    def __init__(
        self,
        name,
        type,
        display_name=None,
        description=None,
        values=None,
        sorter=None,
        meta=None,
        **kwargs
    ):
        if values and isinstance(values, list):
            self.values = set(self.values)

        super(Dimension, self).__init__(
            name,
            type,
            display_name=display_name,
            description=description,
            values=values,
            sorter=sorter,
            meta=meta,
            **kwargs
        )

    def get_values(self, warehouse_id):
        """Get allowed values for this Dimension
        
        **Parameters:**
        
        * **warehouse_id** - (*int*) A zillion warehouse ID
        
        **Returns:**
        
        (*list or None*) - A list of valid values or None if no value
        restrictions have been set.
        
        """
        if self.values is None:
            return None
        if isinstance(self.values, str):
            func = import_object(self.values)
            return func(warehouse_id, self)
        return self.values

    def is_valid_value(self, warehouse_id, value, ignore_none=True):
        """Check if a value is allowed for this Dimension
        
        **Parameters:**
        
        * **warehouse_id** - (*int*) A zillion warehouse ID
        * **value** - (*any*) Check if this value is valid
        * **ignore_none** - (*bool*) If True, consider value=None
        to always be valid.
         
        **Returns:**
        
        (*bool*) - True if the dimension value is valid
        
        """
        if ignore_none and value is None:
            return True
        values = self.get_values(warehouse_id)
        if not values:
            return True
        return value in values

    def sort(self, warehouse_id, values):
        """Sort the given dimension values according to the sorter
        
        **Parameters:**
        
        * **warehouse_id** - (*int*) A zillion warehouse ID
        * **values** - (*Series*) A pandas Series of values to sort
        
        **Returns:**
        
        (*Series*) - A pandas Series representing the sort order
        
        """
        raiseifnot(self.sorter, "No sorter defined on Dimension")
        func = import_object(self.sorter)
        return func(warehouse_id, self, values)


class FormulaField(Field):
    """A field defined by a formula
    
    **Parameters:**
    
    * **name** - (*str*) The name of the field
    * **formula** - (*str*) The formula used to calculate the field
    * **kwargs** - kwargs passed to the super class
    
    """

    repr_attrs = ["name", "formula"]
    schema = FormulaFieldConfigSchema

    def __init__(self, name, formula, **kwargs):
        super(FormulaField, self).__init__(name, None, formula=formula, **kwargs)

    def get_formula_fields(self, warehouse, depth=0, adhoc_fms=None):
        """Get the fields that are part of this field's formula
        
        **Parameters:**
        
        * **warehouse** - (*Warehouse*) A zillion warehouse that will contain
        all relevant fields
        * **depth** - (*int, optional*) Track the depth of recursion into the
        formula
        * **adhoc_fms** - (*list, optional*) A list of FieldManagers
        
        **Returns:**
        
        (*set, str*) - The set of all base fields involved in the formula
        calculation, as well as an expanded version of the formula. All fields
        in the expanded formula should be raw fields (i.e. not formula
        fields).
        
        """
        if depth > MAX_FORMULA_DEPTH:
            raise MaxFormulaDepthException

        raw_formula = self.formula
        raw_fields = set()
        formula_fields = get_string_format_args(self.formula)
        field_formula_map = {}

        for field_name in formula_fields:
            field = warehouse.get_field(field_name, adhoc_fms=adhoc_fms)

            if getattr(field, "technical", None):
                raise InvalidFieldException(
                    "Formula field %s contains field with technical: %s"
                    % (self.name, field.name)
                )

            if isinstance(field, FormulaField):
                try:
                    sub_fields, sub_formula = field.get_formula_fields(
                        warehouse, depth=depth + 1, adhoc_fms=adhoc_fms
                    )
                except MaxFormulaDepthException:
                    if depth != 0:
                        raise
                    raise MaxFormulaDepthException(
                        "Maximum formula recursion depth exceeded for %s: %s"
                        % (self.name, self.formula)
                    )
                for sub_field in sub_fields:
                    raw_fields.add(sub_field)
                field_formula_map[field_name] = "(" + sub_formula + ")"
            else:
                field_formula_map[field_name] = "{" + field_name + "}"
                raw_fields.add(field_name)

        raw_formula = self.formula.format(**field_formula_map)
        return raw_fields, raw_formula

    def get_ds_expression(self, *args, **kwargs):
        """Raise an error if called on FormulaFields"""
        raise ZillionException("Formula-based Fields do not support get_ds_expression")

    def get_final_select_clause(self, warehouse, adhoc_fms=None):
        """Get a SQL select clause for this formula
        
        **Parameters:**
        
        * **warehouse** - (*Warehouse*) A zillion warehouse that will contain
        all relevant fields
        * **adhoc_fms** - (*list, optional*) A list of FieldManagers
        
        **Returns:**
        
        (*SQLAlchemy clause*) - A compiled sqlalchemy clause for the formula
        
        """
        formula_fields, raw_formula = self.get_formula_fields(
            warehouse, adhoc_fms=adhoc_fms
        )
        format_args = {k: k for k in formula_fields}
        formula = raw_formula.format(**format_args)
        if contains_sql_keywords(formula):
            raise DisallowedSQLException(
                "Formula contains disallowed sql: %s" % formula
            )
        return sqla_compile(sa.text(formula))

    def _check_formula_fields(self, warehouse, adhoc_fms=None):
        """Check that all underlying fields exist in the warehouse"""
        fields, _ = self.get_formula_fields(warehouse, adhoc_fms=adhoc_fms)
        if not fields:
            raise InvalidFieldException(
                "No fields found in formula for field:%s formula:%s"
                % (self.name, self.formula)
            )
        for field in fields:
            warehouse.get_field(field, adhoc_fms=adhoc_fms)


class FormulaDimension(FormulaField):
    """A dimension defined by a formula
    
    **Parameters:**
    
    * **name** - (*str*) The name of the dimension
    * **formula** - (*str*) The formula used to calculate the dimension
    * **kwargs** - kwargs passed to super class
    
    """

    field_type = FieldTypes.DIMENSION

    def __init__(self, name, formula, **kwargs):
        # super(FormulaDimension, self).__init__(
        #     name,
        #     formula,
        #     **kwargs
        # )
        raise InvalidFieldException("FormulaDimensions are not currently supported")


class FormulaMetric(FormulaField):
    """A metric defined by a formula
    
    **Parameters:**
    
    * **name** - (*str*) The name of the metric
    * **formula** - (*str*) The formula used to calculate the metric
    * **display_name** - (*str, optional*) The display name of the field
    * **description** - (*str, optional*) The description of the field
    * **meta** - (*dict, optional*) A dict of additional custom attributes
    * **aggregation** - (*str, optional*) The AggregationType to apply to the
    metric
    * **rounding** - (*int, optional*) If specified, the number of decimal
    places to round to
    * **weighting_metric** - (*str, optional*) A reference to a metric to use
    for weighting when aggregating averages
    * **technical** - (*object, optional*) A Technical object or definition used
    to defined a technical computation to be applied to the metric
    * **required_grain** - (*list of str, optional*) If specified, a list of
    dimensions that must be present in the dimension grain of any report that
    aims to include this metric.
    * **kwargs** - kwargs passed to super class
    
    """

    repr_attrs = ["name", "formula", "aggregation", "technical"]
    field_type = FieldTypes.METRIC
    schema = FormulaMetricConfigSchema

    def __init__(
        self,
        name,
        formula,
        display_name=None,
        description=None,
        meta=None,
        aggregation=AggregationTypes.SUM,
        rounding=None,
        weighting_metric=None,
        technical=None,
        required_grain=None,
        **kwargs
    ):
        if technical:
            technical = create_technical(technical)

        super(FormulaMetric, self).__init__(
            name,
            formula,
            display_name=display_name,
            description=description,
            meta=meta,
            aggregation=aggregation,
            rounding=rounding,
            weighting_metric=weighting_metric,
            technical=technical,
            required_grain=required_grain,
            **kwargs
        )


class AdHocField(FormulaField):
    """An AdHoc representation of a field"""

    @classmethod
    def create(cls, obj):
        """Copy this AdHocField"""
        schema = AdHocFieldSchema()
        field_def = schema.load(obj)
        return cls(field_def["name"], field_def["formula"])


class AdHocMetric(FormulaMetric):
    """An AdHoc representation of a Metric

    **Parameters:**

    * **name** - (*str*) The name of the metric
    * **formula** - (*str*) The formula used to calculate the metric
    * **display_name** - (*str, optional*) The display name of the field
    * **description** - (*str, optional*) The description of the field
    * **meta** - (*dict, optional*) A dict of additional custom attributes
    * **technical** - (*object, optional*) A Technical object or definition
    used to defined a technical computation to be applied to the metric
    * **rounding** - (*int, optional*) If specified, the number of decimal
    places to round to
    * **required_grain** - (*list of str, optional*) If specified, a list of
    dimensions that must be present in the dimension grain of any report
    that aims to include this metric.

    """

    schema = AdHocMetricSchema

    def __init__(
        self,
        name,
        formula,
        display_name=None,
        description=None,
        meta=None,
        technical=None,
        rounding=None,
        required_grain=None,
    ):
        """Init an AdHoc representation of a Metric"""
        super(AdHocMetric, self).__init__(
            name,
            formula,
            display_name=display_name,
            description=description,
            meta=meta,
            technical=technical,
            rounding=rounding,
            required_grain=required_grain,
        )

    @classmethod
    def create(cls, obj):
        """Create an AdHocMetric from an AdHocMetricSchema dict"""
        schema = AdHocMetricSchema()
        field_def = schema.load(obj)
        return cls(
            field_def["name"],
            field_def["formula"],
            display_name=field_def["display_name"],
            description=field_def["description"],
            meta=field_def["meta"],
            technical=field_def["technical"],
            rounding=field_def["rounding"],
            required_grain=field_def["required_grain"],
        )


class AdHocDimension(AdHocField):
    """An AdHoc representation of a Dimension"""

    field_type = FieldTypes.DIMENSION


def create_metric(metric_def):
    """Create a Metric object from a dict of params
    
    **Parameters:**
    
    * **metric_def** - (*dict*) A dict of params to init a metric. If a formula
    param is present a FormulaMetric will be created.
    
    """
    if "formula" in metric_def:
        metric = FormulaMetric.from_config(metric_def)
    else:
        metric = Metric.from_config(metric_def)
    return metric


def create_dimension(dim_def):
    """Create a Dimension object from a dict of params
    
    **Parameters:**
    
    * **dim_def** - (*dict*) A dict of params to init a Dimension
    
    """
    if "formula" in dim_def:
        raise InvalidFieldException("FormulaDimensions are not currently supported")
    return Dimension.from_config(dim_def)


class FieldManagerMixin:
    """An interface for managing fields (metrics and dimensions) stored on an
    object.
    
    **Attributes:**
    
    * **metrics_attr** - (*str*) The name of the attribute where metrics are
    stored
    * **dimensions_attr** - (*str*) The name of the attribute where dimensions
    are stored
    
    """

    metrics_attr = "_metrics"
    dimensions_attr = "_dimensions"

    def get_child_field_managers(self):
        """Get a list of child FieldManagers"""
        return []

    def get_field_managers(self, adhoc_fms=None):
        """Get a list of all child FieldManagers including adhoc"""
        return self.get_child_field_managers() + (adhoc_fms or [])

    def get_direct_metrics(self):
        """Get metrics directly stored on this FieldManager"""
        return getattr(self, self.metrics_attr)

    def get_direct_dimensions(self):
        """Get dimensions directly stored on this FieldManager"""
        return getattr(self, self.dimensions_attr)

    def directly_has_metric(self, name):
        """Check if this FieldManager directly stores this metric"""
        return name in getattr(self, self.metrics_attr)

    def directly_has_dimension(self, name):
        """Check if this FieldManager directly stores this dimension"""
        return name in getattr(self, self.dimensions_attr)

    def directly_has_field(self, name):
        """Check if this FieldManager directly stores this field"""
        return name in getattr(self, self.metrics_attr) or name in getattr(
            self, self.dimensions_attr
        )

    def print_metrics(self, indent=None):
        """Print all metrics in this FieldManager"""
        print(format_msg(getattr(self, self.metrics_attr), label=None, indent=indent))

    def print_dimensions(self, indent=None):
        """Print all dimensions in this FieldManager"""
        print(
            format_msg(getattr(self, self.dimensions_attr), label=None, indent=indent)
        )

    def has_metric(self, name, adhoc_fms=None):
        """Check whether a metric is contained in this FieldManager"""
        if self.directly_has_metric(name):
            return True
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            if fm.has_metric(name):
                return True
        return False

    def has_dimension(self, name, adhoc_fms=None):
        """Check whether a dimension is contained in this FieldManager"""
        if self.directly_has_dimension(name):
            return True
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            if fm.has_dimension(name):
                return True
        return False

    def has_field(self, name, adhoc_fms=None):
        """Check whether a field is contained in this FieldManager"""
        if self.directly_has_field(name):
            return True
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            if fm.has_field(name):
                return True
        return False

    def get_metric(self, obj, adhoc_fms=None):
        """Get a reference to a metric on this FieldManager. If the object
        passed is a dict it is expected to defined an AdHocMetric."""
        if isinstance(obj, str):
            if self.directly_has_metric(obj):
                return getattr(self, self.metrics_attr)[obj]
            for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
                if fm.has_metric(obj):
                    return fm.get_metric(obj)
            raise InvalidFieldException("Invalid metric name: %s" % obj)

        if isinstance(obj, dict):
            metric = AdHocMetric.create(obj)
            raiseif(
                self.has_metric(metric.name, adhoc_fms=adhoc_fms),
                "AdHocMetric can not use name of an existing metric: %s" % metric.name,
            )
            metric._check_formula_fields(self, adhoc_fms=adhoc_fms)
            return metric

        raise InvalidFieldException("Invalid metric object: %s" % obj)

    def get_dimension(self, obj, adhoc_fms=None):
        """Get a reference to a dimension on this FieldManager"""
        if isinstance(obj, str):
            if self.directly_has_dimension(obj):
                return getattr(self, self.dimensions_attr)[obj]
            for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
                if fm.has_dimension(obj):
                    return fm.get_dimension(obj)
            raise InvalidFieldException("Invalid dimension name: %s" % obj)

        if isinstance(obj, dict):
            raise InvalidFieldException("AdHocDimensions are not currently supported")

        raise InvalidFieldException("Invalid dimension object: %s" % obj)

    def get_field(self, obj, adhoc_fms=None):
        """Get a refence to a field on this FieldManager"""
        if isinstance(obj, str):
            if self.has_metric(obj, adhoc_fms=adhoc_fms):
                return self.get_metric(obj, adhoc_fms=adhoc_fms)
            if self.has_dimension(obj, adhoc_fms=adhoc_fms):
                return self.get_dimension(obj, adhoc_fms=adhoc_fms)
            raise InvalidFieldException("Invalid field name: %s" % obj)

        if isinstance(obj, dict):
            raise InvalidFieldException("AdHocFields are not currently supported")

        raise InvalidFieldException("Invalid field object: %s" % obj)

    def get_field_instances(self, field, adhoc_fms=None):
        """Get a dict of FieldManagers (including child and adhoc FMs) that
        support a field"""
        instances = {}

        if self.directly_has_field(field):
            instances[self] = self.get_field(field)

        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            if fm.has_field(field):
                instances.update(fm.get_field_instances(field))

        if not instances:
            raise InvalidFieldException("Invalid field name: %s" % field)
        return instances

    def get_metrics(self, adhoc_fms=None):
        """Get a dict of all metrics supported by this FieldManager"""
        metrics = {}
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            fm_metrics = fm.get_metrics()
            metrics.update(fm_metrics)
        metrics.update(getattr(self, self.metrics_attr))
        return metrics

    def get_dimensions(self, adhoc_fms=None):
        """Get a dict of all dimensions supported by this FieldManager"""
        dimensions = {}
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            fm_dimensions = fm.get_dimensions()
            dimensions.update(fm_dimensions)
        dimensions.update(getattr(self, self.dimensions_attr))
        return dimensions

    def get_fields(self, adhoc_fms=None):
        """Get a dict of all fields supported by this FieldManager"""
        fields = {}
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            fm_fields = fm.get_fields()
            fields.update(fm_fields)
        fields.update(getattr(self, self.metrics_attr))
        fields.update(getattr(self, self.dimensions_attr))
        return fields

    def get_direct_fields(self):
        """Get a dict of all fields directly supported by this FieldManager"""
        fields = {}
        fields.update(getattr(self, self.metrics_attr))
        fields.update(getattr(self, self.dimensions_attr))
        return fields

    def get_direct_metric_configs(self):
        """Get a dict of metric configs directly supported by this
        FieldManager"""
        return {f.name: f.to_config() for f in self.get_direct_metrics().values()}

    def get_direct_dimension_configs(self):
        """Get a dict of dimension configs directly supported by this
        FieldManager"""
        return {f.name: f.to_config() for f in self.get_direct_dimensions().values()}

    def get_metric_configs(self, adhoc_fms=None):
        """Get a dict of all metric configs supported by this FieldManager"""
        configs = {}
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            fm_configs = fm.get_metric_configs()
            configs.update(fm_configs)
        configs.update(self.get_direct_metric_configs())
        return configs

    def get_dimension_configs(self, adhoc_fms=None):
        """Get a dict of all dimension configs supported by this FieldManager"""
        configs = {}
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            fm_configs = fm.get_dimension_configs()
            configs.update(fm_configs)
        configs.update(self.get_direct_dimension_configs())
        return configs

    def get_metric_names(self, adhoc_fms=None):
        """Get a set of metric names supported by this FieldManager"""
        return set(self.get_metrics(adhoc_fms=adhoc_fms).keys())

    def get_dimension_names(self, adhoc_fms=None):
        """Get a set of dimension names supported by this FieldManager"""
        return set(self.get_dimensions(adhoc_fms=adhoc_fms).keys())

    def get_field_names(self, adhoc_fms=None):
        """Get a set of field names supported by this FieldManager"""
        return set(self.get_fields(adhoc_fms=adhoc_fms).keys())

    def add_metric(self, metric, force=False):
        """Add a reference to a metric to this FieldManager"""
        if self.has_dimension(metric.name):
            raise InvalidFieldException(
                "Trying to add metric with same name as a dimension: %s" % metric.name
            )
        if (not force) and self.has_metric(metric.name):
            warn("Metric %s already exists on %s" % (metric.name, self))
            return
        getattr(self, self.metrics_attr)[metric.name] = metric

    def add_dimension(self, dimension, force=False):
        """Add a reference to a dimension to this FieldManager"""
        if self.has_metric(dimension.name):
            raise InvalidFieldException(
                "Trying to add dimension with same name as a metric: %s"
                % dimension.name
            )
        if (not force) and self.has_dimension(dimension.name):
            warn("Dimension %s already exists on %s" % (dimension.name, self))
            return
        getattr(self, self.dimensions_attr)[dimension.name] = dimension

    def _add_default_display_names(self, adhoc_fms=None, display_names=None):
        """Populate default display names on all fields"""
        fields = self.get_direct_fields()
        display_names = display_names or {}
        for field, config in fields.items():
            if display_names.get(field, None) and not config.display_name:
                # Make sure we use consistent names if one was specified
                # in a parent field manager.
                config.display_name = display_names[field]
            else:
                default = default_field_display_name(field)
                config.display_name = config.display_name or default
                display_names[field] = config.display_name

        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            fm._add_default_display_names(display_names=display_names)

    def _populate_global_fields(self, config, force=False):
        """Populate fields on this FieldManager from a config
        
        **Parameters:**
        
        * **config** - (*dict*) A config containing lists of metrics and/or
        dimensions to add to this FieldManager
        * **force** - (*bool, optional*) If true, overwrite fields that already
        exist
        
        """
        formula_metrics = []
        formula_dims = []

        for metric_def in config.get("metrics", []):
            if isinstance(metric_def, dict):
                metric = create_metric(metric_def)
            else:
                raiseifnot(
                    isinstance(metric_def, (Metric, FormulaMetric)),
                    "Metric definition must be a dict-like object or a Metric object",
                )
                metric = metric_def

            if isinstance(metric, FormulaMetric):
                formula_metrics.append(metric)  # These get added later
            else:
                self.add_metric(metric, force=force)

        for dim_def in config.get("dimensions", []):
            if isinstance(dim_def, dict):
                dim = create_dimension(dim_def)
            else:
                raiseifnot(
                    isinstance(dim_def, Dimension),
                    "Dimension definition must be a dict-like object or a Dimension object",
                )
                dim = dim_def

            if isinstance(dim, FormulaDimension):
                # formula_dims.append(dim) # These get added later
                raise InvalidFieldException(
                    "FormulaDimensions are not currently supported"
                )
            self.add_dimension(dim, force=force)

        # Defer formulas so params can be checked against existing fields
        for metric in formula_metrics:
            metric._check_formula_fields(self)
            self.add_metric(metric, force=force)

        for dim in formula_dims:
            dim._check_formula_fields(self)
            self.add_dimension(dim, force=force)

    def _find_field_sources(self, field, adhoc_fms=None):
        """Get a list of FieldManagers supporting a field. This will search the
        current FieldManager and all child/adhoc FMs.
        
        **Parameters:**
        
        * **field** - (*str*) The name of a field
        * **adhoc_fms** - (*list, optional*) A list of FieldManagers
        
        **Returns:**
        
        (*list*) - A list of FieldManagers that support the field
        
        """
        sources = []
        if self.directly_has_field(field):
            sources.append(self)

        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            if fm.directly_has_field(field):
                sources.append(fm)
        return sources


def get_table_metrics(fm, table, adhoc_fms=None):
    """Get a list of metrics supported by a table
    
    **Parameters:**
    
    * **fm** - (*FieldManager*) An object supporting the FieldManager interface
    * **table** - (*SQLAlchemy Table*) The table to get a list of supported
    dimensions for
    * **adhoc_fms** - (*list, optional*) AdHoc FieldManagers relevant to this
    request
    
    **Returns:**
    
    (*set*) - A set of metric names
    
    """
    metrics = set()
    for col in table.c:
        if not is_active(col):
            continue
        for field in col.zillion.get_field_names():
            if fm.has_metric(field, adhoc_fms=adhoc_fms):
                metrics.add(field)
    return metrics


def get_table_dimensions(fm, table, adhoc_fms=None):
    """Get a list of dimensions supported by a table
    
    **Parameters:**
    
    * **fm** - (*FieldManager*) An object supporting the FieldManager interface
    * **table** - (*SQLAlchemy Table*) The table to get a list of supported
    dimensions for
    * **adhoc_fms** - (*list, optional*) AdHoc FieldManagers relevant to this
    request
    
    **Returns:**
    
    (*set*) - A set of dimension names
    
    """
    dims = set()
    for col in table.c:
        if not is_active(col):
            continue
        for field in col.zillion.get_field_names():
            if fm.has_dimension(field, adhoc_fms=adhoc_fms):
                dims.add(field)
    return dims


def get_table_fields(table):
    """Get a list of field names supported by a table
    
    **Parameters:**
    
    * **table** - (*SQLAlchemy Table*) The table to get a list of supported
    fields for
    
    **Returns:**
    
    (*set*) - A set of field names
    
    """
    fields = set()
    for col in table.c:
        if not is_active(col):
            continue
        for field in col.zillion.get_field_names():
            fields.add(field)
    return fields


def get_table_field_column(table, field_name):
    """Return the column within a table that supports a given field
    
    **Parameters:**
    
    * **table** - (*Table*) SQLAlchemy table onject
    * **field_name** - (*str*) The name of a field supported by the table
    
    **Returns:**
    
    (*Column*) - A SQLAlchemy column object
    
    """
    for col in table.c:
        if not is_active(col):
            continue
        for field in col.zillion.get_field_names():
            if field == field_name:
                return col
    raise ZillionException(
        "Field %s inactive or not found in table %s" % (field_name, table.fullname)
    )


def table_field_allows_grain(table, field, grain):
    """Check whether a field in a table is restricted by required_grain
    
    **Parameters:**
    
    * **table** - (*Table*) SQLAlchemy table object
    * **field** - (*str*) The name of a field in the table
    * **grain** - (*list of str*) A list of dimenssions that form the target
    grain
    
    """
    grain = grain or set()
    column = get_table_field_column(table, field)
    if not column.zillion.required_grain:
        return True
    if set(column.zillion.required_grain).issubset(grain):
        return True
    return False


def values_from_db(warehouse_id, field):
    """Get allowed field values from the dimension_values table. If
    warehouse_id is `None` the warehouse_id is defaulted to the value
    of `zillion.field.FIELD_VALUE_DEFAULT_WAREHOUSE_ID`. This allows
    pulling dimension values even when a `Warehouse` has not been saved.
    
    **Parameters:**
    
    * **warehouse_id** - (*int*) A zillion warehouse ID
    * **field** - (*Field*) A zillion Dimension object

    **Returns:**
    
    (*list or None*) - A list of valid values or None if no row
    is found for this dimension.
    
    """
    if warehouse_id is None:
        warehouse_id = FIELD_VALUE_DEFAULT_WAREHOUSE_ID

    s = sa.select(DimensionValues.c).where(
        sa.and_(
            DimensionValues.c.warehouse_id == warehouse_id,
            DimensionValues.c.name == field.name,
        )
    )
    conn = zillion_engine.connect()
    try:
        result = conn.execute(s)
        row = result.fetchone()
        if not row:
            return None
        return json.loads(row["values"])
    finally:
        conn.close()


def sort_by_value_order(warehouse_id, field, values):
    """Sort values by the order of the value list defined on the field
    
    **Parameters:**
    
    * **warehouse_id** - (*int*) A zillion warehouse ID
    * **field** - (*Field*) A zillion Field object
    * **values** - (*Series*) A pandas Series to sort
    
    **Returns:**
    
    (*Series*) - A pandas Series representing the sort order. If no value
    list is found for the field, the input values are returned as is.
        
    """
    value_order = field.get_values(warehouse_id)
    if not value_order:
        return values
    mapping = {value: order for order, value in enumerate(value_order)}
    return values.map(mapping)


def get_conversions_for_type(coltype):
    """Get all conversions for a particular column type
    
    **Parameters:**
    
    * **coltype** - A SQLAlchemy column type class
    
    **Returns:**
    
    (*dict*) - The conversion map for the given column type. Returns None if
    no conversions are found.
    
    """
    for basetype, fields in TYPE_ALLOWED_CONVERSIONS.items():
        if issubclass(coltype, basetype):
            return fields
    return None


def replace_non_named_formula_args(formula, column):
    """Do formula arg replacement but raise an error if any named args are present"""
    format_args = get_string_format_args(formula)
    raiseif(
        any([x != "" for x in format_args]),
        "Formula has unexpected named format arguments: %s" % formula,
    )
    if format_args:
        formula = formula.format(*[column_fullname(column) for i in format_args])
    return formula


def get_dialect_type_conversions(dialect, column):
    """Get all conversions supported by this column type for this dialect
    
    **Parameters:**
    
    * **dialect** - (*str*) SQLAlchemy dialect name
    * **column** - (*Column*) SQLAlchemy column object
    
    **Returns:**
    
    (*list*) - A list of dicts containing datasource formulas and criteria
    conversions for each field this column can be converted to
    
    """
    coltype = type(column.type)
    conv_fields = get_conversions_for_type(coltype)
    if not conv_fields:
        return []

    results = []
    for field in conv_fields:
        field_name = field.name
        dialect_field_convs = DIALECT_CONVERSIONS[dialect].get(field_name, None)
        if not dialect_field_convs:
            continue

        if isinstance(dialect_field_convs, str):
            ds_formula = dialect_field_convs
            ds_criteria_conversions = None
        else:
            ds_formula = dialect_field_convs.get("ds_formula", None)
            ds_criteria_conversions = dialect_field_convs.get(
                "ds_criteria_conversions", None
            )

        raiseifnot(
            ds_formula or ds_criteria_conversions,
            "One of ds_formula or ds_criteria_conversions must be set on dialect conversions for %s/%s"
            % (dialect, field_name),
        )

        if ds_formula:
            ds_formula = replace_non_named_formula_args(ds_formula, column)

        results.append(
            dict(
                field=field,
                ds_formula=ds_formula,
                ds_criteria_conversions=ds_criteria_conversions,
            )
        )

    return results


DATETIME_CONVERSION_FIELDS = [
    Dimension("year", "Integer", description="Year"),
    Dimension("quarter", "String(8)", description="Year and quarter (YYYY-QN)"),
    Dimension(
        "quarter_of_year", "SmallInteger", description="Numeric quarter of the year"
    ),
    Dimension("month", "String(8)", description="Year and month (YYYY-MM)"),
    Dimension(
        "month_name",
        "String(8)",
        description="Full name of the month",
        values=[
            "January",
            "February",
            "March",
            "April",
            "May",
            "June",
            "July",
            "August",
            "September",
            "October",
            "November",
            "December",
        ],
        sorter="zillion.field.sort_by_value_order",
    ),
    Dimension("month_of_year", "SmallInteger", description="Numeric month of the year"),
    Dimension("date", "String(10)", description="Date string formatted YYYY-MM-DD"),
    Dimension(
        "day_name",
        "String(10)",
        description="Full name of a day of the week",
        values=[
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ],
        sorter="zillion.field.sort_by_value_order",
    ),
    Dimension(
        "day_of_week",
        "SmallInteger",
        description="Numeric day of the week (monday = 1)",
    ),
    Dimension("day_of_month", "SmallInteger", description="Numeric day of the month"),
    Dimension("day_of_year", "SmallInteger", description="Numeric day of the year"),
    Dimension(
        "hour",
        "String(20)",
        description="Datetime string rounded to the hour (YYYY-MM-DD HH:00:00)",
    ),
    Dimension("hour_of_day", "SmallInteger", description="Numeric hour of day (0-23)"),
    Dimension(
        "minute",
        "String(20)",
        description="Datetime string rounded to the minute (YYYY-MM-DD HH:MM:00)",
    ),
    Dimension(
        "minute_of_hour", "SmallInteger", description="Numeric minute of the hour"
    ),
    Dimension(
        "datetime",
        "String(20)",
        description="Datetime string formatted YYYY-MM-DD HH:MM:SS",
    ),
    # TODO: not fully tested/supported yet, ignore for now
    # Dimension("unixtime", "BigInteger", description="Unix time in seconds"),
]

DATE_CONVERSION_FIELDS = []
for _dim in DATETIME_CONVERSION_FIELDS:
    if _dim.name == "hour":
        break
    DATE_CONVERSION_FIELDS.append(_dim)

# Map all dialect-specific type conversions, including ds_formulas that can
# be applied directly to columns (i.e. DATE(some_datetime) == '2020-01-01')
# and ds_criteria_conversions that can apply conversions to criteria values
# instead.
#
# Note: For date types: somewhat adhering to ISO 8601, but ignoring the "T"
# between the date/time and not including timezone offsets for now because
# zillion assumes everything is in the same timezone (or the datasource
# formulas take care of aligning timezones).
DIALECT_CONVERSIONS = {
    "sqlite": SQLITE_DIALECT_CONVERSIONS,
    "mysql": MYSQL_DIALECT_CONVERSIONS,
    "postgresql": POSTGRESQL_DIALECT_CONVERSIONS,
}

TYPE_ALLOWED_CONVERSIONS = {
    sa.DateTime: DATETIME_CONVERSION_FIELDS,
    sa.DATETIME: DATETIME_CONVERSION_FIELDS,
    sa.TIMESTAMP: DATETIME_CONVERSION_FIELDS,
    sa.Date: DATE_CONVERSION_FIELDS,
    sa.DATE: DATE_CONVERSION_FIELDS,
}



import ast
import os
import re

import sqlalchemy as sa
from sqlalchemy.dialects.mysql import dialect as mysql_dialect
from sqlalchemy.dialects.postgresql import dialect as postgresql_dialect
from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
from sqlalchemy.engine import reflection
from sqlalchemy.engine.url import make_url
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression as exp
import sqlparse as sp

from zillion.core import *

DIGIT_THRESHOLD_FOR_MEAN_AGGR = 1

INTEGER_SA_TYPES = [
    sa.BigInteger,
    sa.BIGINT,
    sa.Integer,
    sa.INT,
    sa.INTEGER,
    sa.SmallInteger,
    sa.SMALLINT,
]

FLOAT_SA_TYPES = [
    sa.DECIMAL,
    sa.Float,
    sa.FLOAT,
    sa.Numeric,
    sa.NUMERIC,
    sa.REAL,
    sa.dialects.postgresql.DOUBLE_PRECISION,
    sa.dialects.postgresql.MONEY,
]

NUMERIC_SA_TYPES = INTEGER_SA_TYPES + FLOAT_SA_TYPES

DATETIME_SA_TYPES = [sa.DateTime, sa.DATETIME, sa.Time, sa.TIME, sa.TIMESTAMP]

DATE_SA_TYPES = [sa.Date, sa.DATE]

AGGREGATION_SQLA_FUNC_MAP = {
    AggregationTypes.MEAN: sa.func.AVG,
    AggregationTypes.COUNT: sa.func.COUNT,
    AggregationTypes.COUNT_DISTINCT: lambda x: sa.func.COUNT(sa.distinct(x)),
    AggregationTypes.MIN: sa.func.MIN,
    AggregationTypes.MAX: sa.func.MAX,
    AggregationTypes.SUM: sa.func.SUM,
}

SQL_AGGREGATION_FUNCS = set(
    [
        "AVG",
        "SUM",
        "MIN",
        "MAX",
        "COUNT",
        "COUNT_DISTINCT",
        "STD",
        "MEDIAN",
        "MODE",
        "VAR",
    ]
)

# This establishes a baseline of schemas to ignore during reflection
DIALECT_IGNORE_SCHEMAS = {
    "mysql": set(["information_schema", "performance_schema", "mysql", "sys"]),
    "postgresql": set(["information_schema", r"pg_(.*)"]),
}


class InvalidSQLAlchemyTypeString(Exception):
    pass


def contains_sql_keywords(sql):
    """Determine whether a SQL query contains special SQL keywords (DML, DDL,
    etc.)
    
    **Parameters:**
    
    * **sql** - (*str or sqlparse result*) The SQL query to check for keywords
    
    **Returns:**
    
    (*bool*) - True if the SQL string contains keywords
    
    """
    if isinstance(sql, str):
        sql = sp.parse(sql)

    for token in sql:
        if token.ttype in (sp.tokens.DML, sp.tokens.DDL, sp.tokens.CTE):
            return True

        if isinstance(token, sp.sql.TokenList):
            token_result = contains_sql_keywords(token)
            if token_result:
                return True
    return False


def contains_aggregation(sql):
    """Determine whether a SQL query contains aggregation functions.
    
    **Warning:**

    This relies on a non-exhaustive list of SQL aggregation functions
    to look for. This will likely need updating.
    
    **Parameters:**
    
    * **sql** - (*str or sqlparse result*) The SQL query to check for
    aggregation functions
    
    **Returns:**
    
    (*bool*) - True if the SQL string contains aggregation
    
    """
    if isinstance(sql, str):
        sql = sp.parse(sql)

    for token in sql:
        if isinstance(token, sp.sql.Function):
            name = token.get_name()
            if name.upper() in SQL_AGGREGATION_FUNCS:
                return True
        if isinstance(token, sp.sql.TokenList):
            token_result = contains_aggregation(token)
            if token_result:
                return True
    return False


def type_string_to_sa_type(type_string):
    """Convert a field type string to a SQLAlchemy type. The type string will be
    evaluated as a python statement or class name to init from the SQLAlchemy
    top level module. Dialect-specific SQLAlchemy types are not currently
    supported.
    
    **Parameters:**
    
    * **type_string** - (*str*) A string representing a SQLAlchemy type, such as
    "Integer", or "String(32)". This does a case-insensitive search and will
    return the first matching SQLAlchemy type.
    
    **Returns:**
    
    (*SQLAlchemy type object*) - An init'd SQLAlchemy type object
    
    """
    try:
        tree = ast.parse(type_string)
        ast_obj = tree.body[0].value
        if isinstance(ast_obj, ast.Name):
            type_name = ast_obj.id
            type_args = []
            type_kwargs = {}
        else:
            type_name = ast_obj.func.id
            type_args = [arg.n for arg in ast_obj.args]
            type_kwargs = {k.arg: k.value.n for k in ast_obj.keywords}

        type_cls = igetattr(sa.types, type_name, None)
        if not type_cls:
            raise InvalidSQLAlchemyTypeString(
                "Could not find matching type for %s" % type_name
            )
        return type_cls(*type_args, **type_kwargs)
    except Exception as e:
        raise InvalidSQLAlchemyTypeString("Unable to parse %s" % type_string) from e


def to_generic_sa_type(type):
    """Return a generic SQLAlchemy type object from a type that may be dialect-
    specific. This will attempt to preserver common type settings such as
    specified field length, scale, and precision. On error it will fall back to
    trying to init the generic type with no params.
    """
    params = {}
    for param in ["length", "precision", "scale"]:
        if hasattr(type, param):
            params[param] = getattr(type, param)
    try:
        return type._type_affinity(**params)
    except Exception as e:
        if "unexpected keyword" not in str(e):
            raise
        return type._type_affinity()


def infer_aggregation_and_rounding(column):
    """Infer the aggregation and rounding settings based on the column type
    
    **Parameters:**
    
    * **column** - (*SQLAlchemy column*) The column to analyze
    
    **Returns:**
    
    (*AggregationType, int*) - A 2-item tuple of the aggregation type and
    rounding to use
    
    """
    if isinstance(column.type, tuple(INTEGER_SA_TYPES)):
        return AggregationTypes.SUM, 0
    if isinstance(column.type, tuple(FLOAT_SA_TYPES)):
        rounding = column.type.scale
        precision = column.type.precision
        if rounding is None and precision is None:
            aggregation = AggregationTypes.SUM
        else:
            whole_digits = precision - rounding
            if whole_digits <= DIGIT_THRESHOLD_FOR_MEAN_AGGR:
                aggregation = AggregationTypes.MEAN
            else:
                aggregation = AggregationTypes.SUM
        return aggregation, rounding
    raise ZillionException("Column %s is not a numeric type" % column)


def aggregation_to_sqla_func(aggregation):
    """Convert an AggregationType string to a SQLAlchemy function"""
    return AGGREGATION_SQLA_FUNC_MAP[aggregation]


def is_numeric_type(type):
    """Determine if this is a numeric SQLAlchemy type"""
    raiseif(isinstance(type, str), "Expected a SQLAlchemy type, got string")
    if isinstance(type, tuple(NUMERIC_SA_TYPES)):
        return True
    return False


def is_probably_metric(column, formula=None):
    """Determine if a column is probably a metric. This is used when trying to
    automatically init/reflect a datasource and determine the field types for
    columns. The logic is very coarse, and should not be relied on for more than
    quick/convenient use cases.
    
    **Parameters:**
    
    * **column** - (*SQLAlchemy column*) The column to analyze
    * **formula** - (*str, optional*) A formula to calculate the column
    
    **Returns:**
    
    (*bool*) - True if the column is probably a metric
    
    """
    if formula and contains_aggregation(formula):
        return True
    if not isinstance(column.type, tuple(NUMERIC_SA_TYPES)):
        return False
    if column.primary_key:
        return False
    if column.name.endswith("_id") or column.name.endswith("Id") or column.name == "id":
        return False
    return True


def sqla_compile(expr):
    """Compile a SQL expression
    
    **Parameters:**
    
    * **expr** - (*SQLAlchemy expression*) The SQLAlchemy expression to compile
    
    **Returns:**
    
    (*str*) - The compiled expression string
    
    """
    return str(expr.compile(compile_kwargs={"literal_binds": True}))


def printexpr(expr):
    """Print a SQLAlchemy expression"""
    print(sqla_compile(expr))


def column_fullname(column, prefix=None):
    """Get a fully qualified name for a column
    
    **Parameters:**
    
    * **column** - (*SQLAlchemy column*) A SQLAlchemy column object to get the
    full name for
    * **prefix** - (*str, optional*) If specified, a manual prefix to prepend to
    the output string. This will automatically be separted with a ".".
    
    **Returns:**
    
    (*str*) - A fully qualified column name. The exact format will vary
    depending on your SQLAlchemy metadata, but an example would be:
    schema.table.column
    
    """
    name = "%s.%s" % (column.table.fullname, column.name)
    if prefix:
        name = prefix + "." + name
    return name


def get_schema_and_table_name(table):
    """Extract the schema and table name from a full table name. If the table
    name is not schema-qualified, return None for the schema name"""
    schema = None
    table_name = table
    if "." in table:
        parts = table.split(".")
        raiseifnot(len(parts) == 2, "Invalid table name: %s" % table)
        schema, table_name = parts
    return schema, table_name


def get_sqla_criterion_expr(column, criterion, negate=False):
    """Create a SQLAlchemy criterion expression
    
    **Parameters:**
    
    * **column** - (*SQLAlchemy column*) A SQLAlchemy column object to be used
    in the expression
    * **criterion** - (*3-item iterable*) A 3-item tuple or list of the format
    [field, operation, value(s)]. See `core.CRITERIA_OPERATIONS` for supported
    operations. The value item may take on different formats depending on the
    operation. In most cases passing an iterable will result in multiple
    criteria of that operation being formed. For example, ("my_field", "=",
    [1,2,3]) would logically or 3 conditions of equality to the 3 values in the
    list. The "between" operations expect each value to be a 2-item iterable
    representing the lower and upper bound of the criterion.
    * **negate** - (*bool, optional*) Negate the expression
    
    **Returns:**
    
    (*SQLAlchemy expression*) - A SQLALchemy expression representing the
    criterion
    
    **Notes:**
    
    Postgresql "like" is case sensitive, but mysql "like" is not. Postgresql
    also supports "ilike" to specify case insensitive, so one option is to look
    at the dialect to determine the function, but that is not supported yet.
    
    """
    field, op, values = criterion
    op = op.lower()
    if not isinstance(values, (list, tuple)):
        values = [values]

    use_or = True
    has_null = any([v is None for v in values])

    if op == "=":
        clauses = [column == v if v is not None else column.is_(None) for v in values]
    elif op == "!=":
        clauses = [column != v if v is not None else column.isnot(None) for v in values]
    elif op == ">":
        clauses = [column > v for v in values]
    elif op == "<":
        clauses = [column < v for v in values]
    elif op == ">=":
        clauses = [column >= v for v in values]
    elif op == "<=":
        clauses = [column <= v for v in values]
    elif op == "in":
        if has_null:
            clauses = [
                column == v if v is not None else column.is_(None) for v in values
            ]
        else:
            clauses = [column.in_(values)]
    elif op == "not in":
        use_or = False
        if has_null:
            clauses = [
                column != v if v is not None else column.isnot(None) for v in values
            ]
        else:
            clauses = [sa.not_(column.in_(values))]
    elif op == "between":
        raiseifnot(len(values) == 2, "Between clause value must have length of 2")
        clauses = [column.between(values[0], values[1])]
    elif op == "not between":
        raiseifnot(len(values) == 2, "Between clause value must have length of 2")
        clauses = [sa.not_(column.between(values[0], values[1]))]
    elif op == "like":
        clauses = [column.like(v) for v in values]
    elif op == "not like":
        use_or = False
        clauses = [sa.not_(column.like(v)) for v in values]
    else:
        raise ZillionException("Invalid criterion operand: %s" % op)

    if use_or:
        clause = sa.or_(*clauses)
    else:
        clause = sa.and_(*clauses)

    if negate:
        clause = sa.not_(clause)

    return clause


def check_metadata_url(url, confirm_exists=False):
    """Check validity of the metadata URL"""
    url = make_url(url)
    dialect = url.get_dialect().name
    if confirm_exists:
        if dialect == "sqlite":
            raiseifnot(
                os.path.isfile(url.database),
                "SQLite DB does not exist: %s" % url.database,
            )
        else:
            raise AssertionError(
                "confirm_exists not supported for dialect: %s" % dialect
            )


def comment(self, c):
    """See https://github.com/sqlalchemy/sqlalchemy/wiki/CompiledComments"""
    self._added_comment = c
    return self


exp.ClauseElement.comment = comment
exp.ClauseElement._added_comment = None


def _compile_element(elem, prepend_newline=False):
    """See https://github.com/sqlalchemy/sqlalchemy/wiki/CompiledComments"""

    @compiles(elem)
    def add_comment(element, compiler, **kw):
        meth = getattr(compiler, "visit_%s" % element.__visit_name__)
        text = meth(element, **kw)
        if element._added_comment:
            # Modified this line to not add newline
            text = "-- %s\n" % element._added_comment + text
        elif prepend_newline:
            text = "\n" + text
        return text


_compile_element(exp.Case)
_compile_element(exp.Label, True)
_compile_element(exp.ColumnClause)
_compile_element(exp.Join)
_compile_element(exp.Select)
_compile_element(exp.Alias)
_compile_element(exp.Exists)


def get_schemas(engine):
    """Inspect the SQLAlchemy engine to get a list of schemas"""
    insp = reflection.Inspector.from_engine(engine)
    return insp.get_schema_names()


# -------- Some DB-specific stuff


def to_mysql_type(type):
    """Compile into a MySQL SQLAlchemy type"""
    return type.compile(dialect=mysql_dialect())


def to_postgresql_type(type):
    """Compile into a PostgreSQL SQLAlchemy type"""
    return type.compile(dialect=postgresql_dialect())


def to_sqlite_type(type):
    """Compile into a SQLite SQLAlchemy type"""
    return type.compile(dialect=sqlite_dialect())


def filter_dialect_schemas(schemas, dialect):
    """Filter out a set of baseline/system schemas for a dialect
    
    **Parameters:**
    
    * **schemas** - (*list*) A list of schema names
    * **dialect** - (*str*) The name of a SQLAlchemy dialect
    
    **Returns:**
    
    (*list*) - A filtered list of schema names
    
    """
    ignores = DIALECT_IGNORE_SCHEMAS.get(dialect, None)
    if not ignores:
        return schemas

    final = []
    for schema in schemas:
        add = True
        for ignore in ignores:
            if re.match(ignore, schema):
                add = False
                break
        if add:
            final.append(schema)

    return final


def get_postgres_schemas(conn):
    """Helper to list PostgreSQL schemas"""
    qr = conn.execute(
        sa.text(
            "SELECT schema_name FROM information_schema.schemata "
            "WHERE schema_name not LIKE 'pg_%' and schema_name != 'information_schema'"
        )
    )
    return [x["schema_name"] for x in qr.fetchall()]


def get_postgres_pid(conn):
    """Helper to get the PostgreSQL connection PID"""
    qr = conn.execute("select pg_backend_pid()")
    pid = qr.fetchone()[0]
    return pid
